#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include "debug.h"
#include "struct.h"
#include "daemon.h"
#include "util.h"
#include "show.h"

int connect_socket (sladd_config_t * config) {
  int sock;
  struct sockaddr_in ipv4 = {
    sin_family: AF_INET,
    sin_port: htons (config->listenport),
    sin_addr: { htonl (INADDR_LOOPBACK) }
  };
  struct sockaddr_in6 ipv6 = {
    sin6_family: AF_INET6,
    sin6_port: htons (config->listenport),
    sin6_flowinfo: 0,
    sin6_addr: IN6ADDR_LOOPBACK_INIT,
    sin6_scope_id: 0
  };

  if ((sock = socket (PF_INET6, SOCK_STREAM, 0)) != -1) {
    if (!connect (sock, (struct sockaddr *)&ipv6, sizeof (ipv6)))
      return sock;
    close (sock);
  }
  if (errno != EAFNOSUPPORT)
    {
      PRINTF("could not create IPv6 socket\n");
      return -1;
    }

  if ((sock = socket (PF_INET, SOCK_STREAM, 0)) != -1) {
    if (!connect (sock, (struct sockaddr *)&ipv4, sizeof (ipv4)))
      return sock;
    close (sock);
  }

  PRINTF("could not create IPv4 socket\n");
  return -1;
}

int send_server (const char * command, sladd_config_t * config) {
  int socket, result, len;
  fd_set set;
  char buf[1024];

  /* Get a connected socket */
  socket = connect_socket (config);
  if (socket == -1)
    {
      PRINTF("send_server(): could not connect_socket()\n");
      return 1;
    }

  /* Send command */
  write (socket, command, strlen (command));

  /* Receive from server, print everything received to stdout*/
  do {
    FD_ZERO (&set);
    FD_SET (socket, &set);
    result = select (socket + 1, &set, NULL, NULL, NULL);
    if (result >= 0)
      if (FD_ISSET (socket, &set)) {
	len = recv (socket, buf, 1023, 0);
	if (len >= 0) {
	  buf[len] = 0;
	  printf ("%s", buf);
	}
      }
  } while (result >= 0 && len > 0);
  shutdown (socket, SHUT_RDWR);
  close (socket);
  return 0;
}

/* Queue a command on the server */
int cl_runcmd (sladd_config_t * config) {
  char command[256];

  snprintf (command, sizeof(command), "run %s\n", config->arg);

  if (send_server (command, config)) {
    /* Send failed, maybe the daemon isn't running */
    if (start_daemon (config))
      {
	PRINTF("could not start the daemon process\n");
	return 1;
      }
    /* Daemon started, retry */
    sleep(1); /* give server time to start up */
    if (send_server (command, config)) {
      fprintf (stderr, MSG_CANTQUEUE);
      return 1;
    }
  }
  return 0;
}

/* Perform one of the show actions */
int cl_show (sladd_config_t * config) {
  char cmd[256];
  int i, coloncount = 0;

  if (!config->arg) {
    fprintf (stderr, "No selection to --show\n");
    return 1;
  }

  /* List plugins */
  if (!strcasecmp (config->arg, "plugins")) {
    if (send_server ("show plugins\n", config)) {
      /*
       * Couldn't list from server, print out own list
       * This is the same routine called from the daemon, using
       * the stdout == fd 1 assumption
       */
      show_plugins (1, config);
    }
  } else if (!strcasecmp (config->arg, "alljobs")) {
    return send_server ("show alljobs\n", config);
  } else if (!strcasecmp (config->arg, "jobs")) {
    send_server ("show jobs\n", config);
  } else {
  /* Couldn't identify request directly, let's see if it's a plugin */
    for (i=0; config->arg[i]; i++)
      if (config->arg[i] == ':')
	coloncount++;
    if (coloncount == 2) {
      snprintf (cmd, 256, "show %s\n", config->arg);
      return send_server (cmd, config);
    } else {
      printf (MSG_NOCOMMAND);
    }
  }
  return 0;
}

/* Externally visible entry point from main () */
int run_client (sladd_config_t * config) {
  switch (config->command) {
  case 'q':
    return send_server ("quit\n", config);
  case 'r':
    return cl_runcmd (config);
  case 's':
    return cl_show (config);
  default:
    fprintf (stderr, "Unknown command '%c'\n", config->command);
    return 1;
  }
  return 0;
}
