#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <signal.h>
#include <pthread.h>

#include "debug.h"
#include "struct.h"
#include "show.h"
#include "util.h"
#include "config.h"
#include "process.h"

int create_socket (sladd_config_t * config) {
  int sock;
  struct sockaddr_in ipv4 = {
    sin_family: AF_INET,
    sin_port: htons (config->listenport),
    sin_addr: { htonl (INADDR_LOOPBACK) }
  };
  struct sockaddr_in6 ipv6 = {
    sin6_family: AF_INET6,
    sin6_port: htons (config->listenport),
    sin6_flowinfo: 0,
    sin6_addr: IN6ADDR_LOOPBACK_INIT,
    sin6_scope_id: 0
  };

  if ((sock = socket (PF_INET6, SOCK_STREAM, 0)) != -1) {
    reuse (sock);
    if (!bind (sock, (struct sockaddr *)&ipv6, sizeof (ipv6)))
      {
	PRINTF("using IPv6\n");
	return sock;
      }
    close (sock);
  }
  if (errno != EAFNOSUPPORT)
    return -1;
  if ((sock = socket (PF_INET, SOCK_STREAM, 0)) != -1) {
    if (sock == -1) {
      perror ("socket()");
      return -1;
    }
    reuse (sock);
    if (!bind (sock, (struct sockaddr *)&ipv4, sizeof (ipv4)))
      {
	PRINTF("using IPv4\n");
	return sock;
      }
    perror ("bind()");
    close (sock);
  }
  return -1;
}

int daemonize () {
  int i, file, result;

  sigset_t signals;
  sigemptyset (&signals);
  sigaddset (&signals, SIGCHLD);
  if (signal (SIGCHLD, SIG_DFL) == SIG_ERR) {
    perror ("could not set SIGCHLD handler to default");
    return -1;
  }
  if (sigprocmask (SIG_UNBLOCK, &signals, NULL)) {
    perror ("sigprocmask ()");
    return -1;
  }

  /* First fork() */
  switch (fork ()) {
  case -1: /* error */
    perror ("fork()");
    return -1;
  case 0:  /* child */
    break;
  default: /* parent */
    return 1;
  }

  /* new process group */
  result = setsid ();
  if (result < 0) {
    perror ("setsid()");
    return result;
  }

  /* Second fork() */
  switch (fork ()) {
  case -1: /* error */
    perror ("fork()");
    return -1;
  case 0:  /* child */
    break;
  default: /* parent */
    exit (0);
  }

  /* Our cwd might disappear */
  chdir ("/");

  /* Reassign stdin/out/err */
  file = open ("/dev/null", O_RDWR);
  if (file < 0)
    return file;

  for (i = 0; i < 3; i++)
    if (file != i) {
      result = dup2 (file, i);
      if (result < 0)
	return result;
    }

  close (file);

  return 0;
}

void action (const char * cmd, int socket, sladd_config_t * config) {
  int len = strlen (cmd);

  PRINTF("action(%s)\n", cmd);

  if (!strncasecmp (cmd, "quit", 4)) {
    write (socket, MSG_SHUTDOWN, strlen (MSG_SHUTDOWN));
    kill (0, SIGTERM);
  } else if (!strncasecmp (cmd, "show", 4)) {
    if (len > 5) {
      cmd+=5;
      if (!strncasecmp (cmd, "plugins", 7)) {
	show_plugins (socket, config);
      } else if (!strncasecmp (cmd, "jobs", 4)) {
	show_jobs (socket);
      } else {
	show_job_output (socket, cmd);
      }
    }
  } else if (!strncasecmp (cmd, "run", 3)) {
    if (len > 4) {
      run_plugin (socket, cmd+4, config);
    }
  }

  return;
}

void handle_socket (receive_buffer_t * buffer, sladd_config_t * config) {
  int len;
  char * tmp;

  len = recv (buffer->socket, buffer->recvbuf + buffer->recvlen,
	      RECVBUFSIZE - 1 - buffer->recvlen, 0);
  if (len > 0) {
    buffer->recvlen += len;
    buffer->recvbuf[buffer->recvlen+1] = 0;
#if 1
    buffer->last_event = time (NULL);
#endif
    if ((tmp = strchr (buffer->recvbuf, '\n'))) {
      *tmp = 0;
      action (buffer->recvbuf, buffer->socket, config);
    }
    if ((tmp && !tmp[0]) || buffer->recvlen == (RECVBUFSIZE - 1)) {
      shutdown (buffer->socket, SHUT_RDWR);
      close (buffer->socket);
      buffer->socket = -1;
    }
  }
}

/*
 * Just to change the return code: SIGTERM to exit(0)
 */
static void terminate (int signal) {
  PRINTF("SIGTERM\n");
  exit (0);
}

/*
 * Entry point for main() and from cl_run()
 */
int start_daemon (sladd_config_t * config) {
  int socket, client;
  receive_buffer_t recvbuf[SLADD_MAX_CONNECTIONS];
  fd_set recv_set;
  int i, max_fd, result;
  pid_t term_child;
  struct timeval tv;
  time_t now;

  /* Start listening */
  socket = create_socket (config);
  if (socket < 0)
    {
      PRINTF("could not create socket\n");
      return 1;
    }
  if (listen (socket, 1))
    perror ("listen()");

  memset (recvbuf, 0, sizeof (recvbuf));

  /* Init signal handlers */
  if (process_init ())
    {
      PRINTF("could not process_init()\n");
      return 1;
    }

  /* Now daemonize */
#ifndef DOJDEBUG
  switch (daemonize ()) {
  case -1: /* Error */
    PRINTF("error in daemonize()\n");
    return 1;
  case 1:  /* Original parent (only returns once */
    return 0;
  }
  /* From now on, we're the double-forked child without stdin/out/err */
#endif

  /* Catch SIGTERM just to change the return code */
  signal (SIGTERM, terminate);

  while (1) {
    /* build a set from all possible active channels */
    FD_ZERO (&recv_set);
    max_fd = get_joblist (&recv_set);   /* The msg sockets from the plugins */
    FD_SET (socket, &recv_set);            /* The inbound connection socket */
    for (i=0; i<SLADD_MAX_CONNECTIONS; i++)
      if (recvbuf[i].socket > 0) {
	FD_SET (recvbuf[i].socket, &recv_set);      /* The command channels */
	if (recvbuf[i].socket > max_fd)
	  max_fd = recvbuf[i].socket;
      }

    if (socket > max_fd)
      max_fd = socket;

    tv.tv_sec = 1;
    tv.tv_usec = 0;
    result = select (max_fd + 1, &recv_set, NULL, NULL, &tv);
    switch (result) {
    case -1:
      if (errno != EINTR)
	{
	  PRINTF("select error errno=%i %s\n", errno, strerror(errno));
	  kill (0, SIGTERM);
	}
      /* probably interrupted by SIGCHLD, let's see... */
      break;
    case 0:
      now = time (NULL);
      for (i=0; i<SLADD_MAX_CONNECTIONS; i++)
	if (recvbuf[i].socket > 0 && ((recvbuf[i].last_event + SLADD_SESSION_TIMEOUT) < now)) {
	  shutdown (recvbuf[i].socket, SHUT_RDWR);
	  close (recvbuf[i].socket);
	  recvbuf[i].socket = -1;
	  recvbuf[i].last_event = 0;
	}
      break;
    default:
      if (FD_ISSET (socket, &recv_set)) {
	if ((client = accept (socket, NULL, NULL)) > 0) {
	  for (i=0; i<SLADD_MAX_CONNECTIONS; i++)
	    if (recvbuf[i].socket <= 0) {
	      recvbuf[i].socket = client;
	      recvbuf[i].recvlen = 0;
	      recvbuf[i].last_event = time (NULL);
	      recvbuf[i].recvbuf[0] = 0;
	      break;
	    }
	  if (i == SLADD_MAX_CONNECTIONS) {
	    write (client, MSG_MAXCONNS, strlen (MSG_MAXCONNS));
	    shutdown (client, SHUT_RDWR);
	    close (client);
	  }
	}
      }
      for (i=0; i<SLADD_MAX_CONNECTIONS; i++)
	if (recvbuf[i].socket > 0 && FD_ISSET (recvbuf[i].socket, &recv_set))
	  handle_socket (&recvbuf[i], config);
      receive (&recv_set);
    }
    if ((term_child = waitpid (-1, NULL, WNOHANG)) > 0)
      reap_job (term_child);
  }
  /* should never get here */
  return 0;
}
