#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <libgen.h>
#include <errno.h>
#include <stdio.h>
#define __GNU_SOURCE
#include "getopt.h"

#include "debug.h"
#include "config.h"
#include "xml.h"
#include "struct.h"

sladd_config_t *  init_defaults () {
  sladd_config_t * config;

  config = calloc (1, sizeof (sladd_config_t));
  if (!config)
    {
      PRINTF("init_default(): could not calloc config\n");
      return NULL;
    }

  config->listenport = SLADD_PORT;
  config->pluginpath = strdup (SLAD_PLUGIN_PATH);

  if (!config->pluginpath) {
    free (config);
    config = NULL;
  }

  return config;
}

void init_environment (sladd_config_t * config) {
  char * envvar;
  int tmp;

  envvar = getenv ("SLADPORT");
  if (envvar) {
    tmp = atoi (envvar);
    if (tmp > 0 && tmp < 65536)
      config->listenport = tmp;
  }
  envvar = getenv ("SLADPATH");
  if (envvar) {
    free (config->pluginpath);
    config->pluginpath = strdup (envvar);
  }
}

void print_help (char * arg0) {
  printf ("Usage: %s [OPTION...]\n", basename (arg0));
  printf ("-p, --pluginpath=%-21s plug-in directory\n", SLAD_PLUGIN_PATH);
  printf ("-l, --listenport=%-21d the port the daemon will listen on\n", SLADD_PORT);
  printf ("-d, --daemonize                        run as daemon\n"
	  "-s, --show=plugins|jobs|<resultid>     display status\n"
	  "-r, --run=pluginid                     run a job\n"
	  "-q, --quit                             terminate the daemon\n"
	  "-h, --help                             display this help\n");

  _exit (0);
}

void init_cmdline (int argc, char * const argv[], sladd_config_t * config) {
  int daemonize = 0, listenport = 0;
  char * show = NULL, * pluginpath = NULL, * command = NULL;
  int i;
  int option_index = 0;

  static const struct option long_options[] = {
    {"pluginpath", 1, NULL, 'p'},
    {"listenport", 1, NULL, 'l'},
    {"daemonize", 0, NULL, 'd'},
    {"show", 1, NULL, 's'},
    {"run", 1, NULL, 'r'},
    {"quit", 0, NULL, 'q'},
    {"help", 0, NULL, 'h'},
    {NULL, 0, NULL, 0}
  };

  while (1) {
    i = getopt_long (argc, argv, "dhl:p:qr:s:", long_options, &option_index);
    if (i == -1)
      break;

    switch (i) {
    case 'd':
      daemonize = 1;
      break;
    case 'h':
      print_help (argv[0]); /* never returns */
      break;
    case 'l':
      listenport = atoi (optarg);
      break;
    case 'p':
      if (pluginpath)
	free (config->pluginpath);
      pluginpath = strdup (optarg);
      break;
    case 'q':
      config->command = 'q';
      break;
    case 'r':
      if (command)
	free (command);
      command = strdup (optarg);
      break;
    case 's':
      if (show)
	free (show);
      show = strdup (optarg);
      break;
    }
  }

  if (pluginpath) {
    free (config->pluginpath);
    config->pluginpath = pluginpath;
  }
  if (listenport > 0 && listenport < 65536)
    config->listenport = listenport;

  if (daemonize)
    config->command = 'd';
  else if (show) {
    config->command = 's';
    config->arg = show;
  } else if (command) {
    config->command = 'r';
    config->arg = command;
  }

  if (!config->command)
    print_help (argv[0]);
}

void init_plugins (sladd_config_t * config) {
  char pluginpath[1024], tmppath[1024];
  struct stat statbuf;
  int result;
  DIR * dir;
  struct dirent * entry;

  strcpy (pluginpath, config->pluginpath);
  do {
    if (lstat (pluginpath, &statbuf) < 0)
      {
	PRINTF("init_plugins(): lstat() error\n");
	return;
      }
    if (S_ISLNK (statbuf.st_mode)) {
      result = readlink (pluginpath, tmppath, 1023);
      if (result >= 0) {
	tmppath[result] = 0;
	if (tmppath[0] == '/')
	  strcpy (pluginpath, tmppath);
	else {
	  strcpy (pluginpath, dirname (pluginpath));
	  strcat (pluginpath, "/");
	  strcat (pluginpath, tmppath);
	}
      }
      else
	{
	  PRINTF("init_plugins(): readlink() error\n");
	  return;
	}
    } else if (!S_ISDIR (statbuf.st_mode))
      {
	PRINTF("init_plugins(): !S_ISDIR()\n");
	return;
      }
  } while (S_ISLNK (statbuf.st_mode));

  if (!(dir = opendir (pluginpath))) {
    fprintf (stderr, "can't opendir() plugin directory %s: %s\n",
	     pluginpath, strerror (errno));
    return;
  }

  while ((entry = readdir (dir))) {
    if (strcmp (entry->d_name, ".") && strcmp (entry->d_name, "..")) {
      snprintf (tmppath, 1024, "%s/%s", pluginpath, entry->d_name);
      if (!lstat (tmppath, &statbuf) && S_ISDIR (statbuf.st_mode)) {
	strcat (tmppath, "/plugin.xml");
	if (!lstat (tmppath, &statbuf) && S_ISREG (statbuf.st_mode))
	  {
	    load_plugin (tmppath, config);
	  }
      }
    }
  }

  if (strcmp (pluginpath, config->pluginpath)) {
    free (config->pluginpath);
    config->pluginpath = strdup (pluginpath);
  }

  closedir (dir);
}

sladd_config_t * init (int argc, char * const argv[]) {
  sladd_config_t * config;

  config = init_defaults ();
  if (config) {
    init_environment (config);
    init_cmdline (argc, argv, config);
    init_plugins (config);
  }

  return config;
}
