#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/select.h>
#include <sys/wait.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <pthread.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>
#include <signal.h>

#include "debug.h"
#include "struct.h"
#include "util.h"

pthread_mutex_t joblock = PTHREAD_MUTEX_INITIALIZER;
sladd_job_t * jobs = NULL;

/*
 * List the jobs as response to "show jobs"
 */
void show_jobs (int socket) {
  sladd_job_t * job;
  char buf[1024];

  for (job = jobs; job; job = job->next) {
    snprintf (buf, 1024, "%c:%s:%d@%p\n",
	      job->running ? 'R' : 'T', job->pluginid, job->outputlen, job);
    write (socket, buf, strlen (buf));
  }
}

/*
 * List the output of a terminated job as response to "show <jobid>"
 */
void show_job_output (int socket, const char * jobid) {
  sladd_job_t * job, * last = NULL;
  int done = 0, result = 0;

  for (job = jobs; job; last = job, job = job->next)
    if (!job->running && !strcasecmp (jobid, job->pluginid)) {
      while (done < job->outputlen && result >= 0) {
	result = write (socket, job->output+done, job->outputlen - done);
	done += result;
      }
      if (last)
	last->next = job->next;
      else
	jobs = job->next;
      free (job->output);
      free (job);
      return;
    }
  write (socket, MSG_NOPLUGIN, strlen (MSG_NOPLUGIN));
}

int get_joblist (fd_set * set) {
  int max_fd = 0;
  sladd_job_t * job;

  for (job = jobs; job; job = job->next)
    if (job->socket > 0) {
      FD_SET (job->socket, set);
      if (job->socket > max_fd)
	max_fd = job->socket;
    }
  return max_fd;
}

/*
 * Actual receiver
 */
void do_receive (sladd_job_t * job) {
  int len;
  char buf[1024];
  char * tmp;

  while ((len = recv (job->socket, buf, sizeof(buf), 0)) > 0) {
    tmp = realloc (job->output, job->outputlen + len);
    if (tmp) {
      job->output = tmp;
      memcpy (job->output+job->outputlen, buf, len);
      job->outputlen += len;
    }
  }
}

void receive (fd_set * set) {
  sladd_job_t * job;

  for (job = jobs; job; job = job->next)
    if (job->socket > 0 && FD_ISSET (job->socket, set))
      do_receive (job);
}

/*
 * Pull the exit status after a SIGCHLD to prevent zombiefication
 * and to update the table of running jobs
 */
void reap_job (pid_t pid) {
  sladd_job_t * job;

  for (job = jobs; job; job = job->next)
    if (job->pid == pid) {
      close (job->socket);
      job->running = 0;
      job->socket = -1;
      job->pid = 0;
      PRINTF("reap_job(%i) success\n", pid);
      break;
    }
}

int process_init () {
  sigset_t signals;

  sigemptyset (&signals);
  sigaddset (&signals, SIGCHLD);
  sigaddset (&signals, SIGCONT);
  sigprocmask (SIG_UNBLOCK, &signals, NULL);

  return 0;
}

#ifdef DOJDEBUG
#define doerr(cmsg) do { PRINTF("%s: %s\n", cmsg, strerror(errno)); return -1; } while(0)
pid_t commfork (int socket, const char * id) {
#else
#define doerr(cmsg) return -1
pid_t commfork (const char * id) {
#endif
  int paired_fds[2] = {-1, -1};
  pid_t pid;
  sladd_job_t * job, * tmp;
  int fileflags;

  if (socketpair (AF_UNIX, SOCK_STREAM, 0, paired_fds) < 0)
    doerr ("socketpair ()");
  if ((fileflags = fcntl (paired_fds[0], F_GETFL, 0)) == -1)
    doerr ("fcntl getfl ()");
  if (fcntl (paired_fds[0], F_SETFL, fileflags | O_NONBLOCK))
    doerr ("fcntl setfl ()");

  job = calloc (sizeof (sladd_job_t), 1);
  if (!job)
    doerr ("calloc ()");

  switch ((pid = fork ())) {
  case 0:
    /* child */

    /* wait for parent to setup their channels */
    kill (getpid (), SIGSTOP);

    /* setup my channels */
    close (paired_fds[0]);
    close (1);
    close (2);
    if (dup2 (paired_fds[1], 1) != 1)
      doerr ("dup2 1");
    if (dup2 (paired_fds[1], 2) != 2)
      doerr ("dup2 2");
    /*
    stdout = fdopen (1, "w");
    stderr = fdopen (2, "w");
    */
    paired_fds[0] = open ("/dev/null", O_RDONLY);
    dup2 (paired_fds[0], 0);
    /*
    stdin = fdopen (0, "r");
    */
    close (paired_fds[0]);
    return 0;
  case -1:
    /* error */
    doerr ("fork ()");
  default:
    /* parent */
    close (paired_fds[1]);
    job->pid = pid;
    job->running = 1;
    job->socket = paired_fds[0];
    job->pluginid = strdup (id);

    if (!jobs) {
      jobs = job;
    } else {
      for (tmp = jobs; tmp->next; tmp = tmp->next);
      tmp->next = job;
    }

    /* signal child that we're ready */
    waitpid (pid, NULL, WUNTRACED);
    if (kill (pid, SIGCONT)) {
      doerr ("kill ()");
    }

    return pid;
  }
}

#define MAX_ARGV      100
#define MAX_MSG       512

int start_plugin (int socket, slad_plugin_t * p, slad_plugin_set_t * s,
		  slad_plugin_entry_t * e, sladd_config_t * config) {
 char path[256];
  char * argv[MAX_ARGV], * cmdline;
  int argc, i;
  char msg[MAX_MSG];
  sladd_job_t * job = NULL;
  struct stat statbuf;
  pid_t pid;
  struct passwd * passwd = NULL;
  struct group * group = NULL;
  char pluginname[1024];

  snprintf (pluginname, 1024, "%s:%s:%s", p->name, s->name, e->name);

  for (job = jobs; job; job = job->next)
    if (!strcasecmp (pluginname, job->pluginid))
      break;

  if (job) {
    write (socket, MSG_ALREADY, strlen (MSG_ALREADY));
    return -1;
  }

  strcpy (path, config->pluginpath);
  if (p->path) {
    strcat (path, "/");
    strcat (path, p->path);
  }
  strcat (path, "/");
  strcat (path, e->binary);

  if (stat (path, &statbuf)) {
    snprintf (msg, MAX_MSG, MSG_CANTRUN " - stat(%s): %s\n",
	      path, strerror (errno));
    write (socket, msg, strlen (msg));
    return -1;
  }
  if (!S_ISREG (statbuf.st_mode)) {
    snprintf (msg, MAX_MSG, MSG_CANTRUN " - %s is not a regular file\n", path);
    write (socket, msg, strlen (msg));
    return -1;
  }
  if (!(statbuf.st_mode & (S_IXUSR | S_IXGRP | S_IXOTH))) {
    snprintf (msg, MAX_MSG, MSG_CANTRUN " - %s is not executable\n", path);
    write (socket, msg, strlen (msg));
    return -1;
  }

  argv[0] = path;
  for (argc = 1, cmdline = e->cmdline; argc < MAX_ARGV - 1; cmdline = NULL)
    if (!(argv[argc++] = strtok (cmdline, " \t\n")))
      break;
  argv[MAX_ARGV-1] = NULL;

  if (e->runasuser)
    passwd = getpwnam (e->runasuser);
  if (e->runasgroup)
    group = getgrnam (e->runasgroup);

#ifdef DOJDEBUG
  switch (pid = commfork (socket, pluginname)) {
#else
  switch (pid = commfork (pluginname)) {
#endif

  case 0:
    if (group)
      setgid (group->gr_gid);
    if (passwd)
      setuid (passwd->pw_uid);
    /* no provileged operations after this point */
    for (i=3; i < FD_SETSIZE; i++)
      close (i);
    PRINTF("execvp(%s) pid=%i\n", argv[0], getpid());
    execvp (argv[0], argv);
    snprintf (msg, MAX_MSG, "execvp(): %s\n", strerror (errno));
    write (socket, msg, strlen (msg));
    _exit(1);
  case -1:
    write (socket, MSG_CANTRUN " - fork error\n", strlen (MSG_CANTRUN)+14);
    return -1;
    break;
  default:
    snprintf (msg, MAX_MSG, "%s %s\n", MSG_OK, p->desc);
    write (socket, msg, strlen (msg));
    return 0;
  }
}

/*
 * Execute the "run <plugin>" command
 */
int run_plugin (int socket, const char * pluginname, sladd_config_t * config) {
  slad_plugin_t * p = NULL;
  slad_plugin_set_t * s = NULL;
  slad_plugin_entry_t * e = NULL;
  int e_num;

  findplugin (pluginname, config, &p, &s, &e);
  if (!p) {
    PRINTF("1: did not find the plugin\n");
    write (socket, MSG_NOPLUGIN, strlen (MSG_NOPLUGIN));
    return -1;
  }
  if (!s)
    s = (p->sets)[0];
  if (!s) {
    PRINTF("2: did not find the plugin\n");
    write (socket, MSG_NOPLUGIN, strlen (MSG_NOPLUGIN));
    return -1;
  }

  if (e)
    return start_plugin (socket, p, s, e, config);

  if (!(s->entries)[0]) {
    PRINTF("3: did not find the plugin\n");
    write (socket, MSG_NOPLUGIN, strlen (MSG_NOPLUGIN));
    return -1;
  }

  for (e_num = 0; (e = s->entries[e_num]); e_num++)
    start_plugin (socket, p, s, e, config);

  return 0;
}
