#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <libgen.h>

#include <expat.h>

#include "debug.h"
#include "struct.h"
#include "util.h"

#ifndef XMLCALL
#define XMLCALL
#endif

int depth;

#define SLAD_NS "slad:"

// This is used for parsing the elements
static const char * elements[] = {
  // slad:Plugin
  "plugin",
  "pluginDescription",
  "pluginPath",

  // slad:PluginSet
  "pluginSet",
  "pluginSetDescription",
  "pluginSetDefault",

  // slad:PluginSetEntry
  "pluginSetEntry",
  "pluginSetEntryDescription",
  "pluginSetEntryDefault",
  "pluginSetEntryBinary",
  "pluginSetEntryCommandline",
  "pluginSetEntryMaxRuntime",
  "pluginSetEntryMaxLoglines",
  "pluginSetEntryRunAsUser",
  "pluginSetEntryRunAsGroup",

  NULL
};

// This must match the string array above
enum SLADTAGS {
  ST_PLUGIN                   = 0,
  ST_PLUGINDESC,
  ST_PLUGINPATH,

  ST_PLUGINSET,
  ST_PLUGINSETDESC,
  ST_PLUGINSETDEFAULT,

  ST_PLUGINSETENTRY,
  ST_PLUGINSETENTRYDESC,
  ST_PLUGINSETENTRYDEFAULT,
  ST_PLUGINSETENTRYBINARY,
  ST_PLUGINSETENTRYCMDLINE,
  ST_PLUGINSETENTRYMAXRUNTIME,
  ST_PLUGINSETENTRYMAXLOGLINES,
  ST_PLUGINSETENTRYRUNASUSER,
  ST_PLUGINSETENTRYRUNASGROUP
};

const char * getattr (const char * name, const char ** attr) {
  int i;

  for (i = 0; attr[i]; i += 2)
    if (!strcmp (attr[i], name))
      return attr[i+1];

  return "";
}

static void XMLCALL nullhandler (void * data, const char * s, int len) {
  //  fprintf (stderr, "XML NULL handler called\n");
}

static void XMLCALL stringhandler (void * data, const char * s, int len) {
  XML_Parser parser = (XML_Parser)data;
  char * text, * tmp;
  char ** target;

  target = (char **) XML_GetUserData (parser);
  if (len > 0) {
    text = tmp = malloc(len+1);
    if (!text)
      {
	PRINTF("xml stringhandler(): could not malloc text\n");
	return;
      }
    strncpy(text, s, len);
    text[len] = 0;

    // strip white spaces at the beginning
    while (*tmp && isspace(*tmp)) {
      tmp++;
      len--;
    }
    // strip white spaces at the end
    while (len && isspace (tmp[len-1])) {
      len--;
      tmp[len] = 0;
    }

    if (len) {
      if (*target)
	free (*target);
      *target = strdup (tmp);
    }
    free (text);
  }
}

static void XMLCALL inthandler (void * data, const char * s, int len) {
  XML_Parser parser = (XML_Parser)data;
  char * text, * tmp;
  int * target;

  target = (int *) XML_GetUserData (parser);
  if (len > 0) {
    text = tmp = malloc(len+1);
    if (!text)
      {
	PRINTF("xml inihandler(): could not malloc text\n");
	return;
      }
    strncpy(text, s, len);
    text[len] = 0;

    // strip white spaces at the beginning
    while (*tmp && isspace(*tmp)) {
      tmp++;
      len--;
    }
    // strip white spaces at the end
    while (len && isspace (tmp[len-1])) {
      len--;
      tmp[len] = 0;
    }

    if (len)
      *target = atoi (tmp);
    free (text);
  }
}

static void XMLCALL start (void * data, const char * el, const char ** attr) {
  XML_Parser parser = (XML_Parser)data;
  static sladd_config_t * config = NULL;
  static slad_plugin_t * plugin = NULL;
  static slad_plugin_set_t * set = NULL;
  static slad_plugin_entry_t * entry = NULL;
  const char * element;
  int i;

  if (!config)
    config = XML_GetUserData (parser);
  if (strncmp (el, SLAD_NS, strlen (SLAD_NS))) {
    fprintf (stderr, "Element %s outside %s namespace", el, SLAD_NS);
    return;
  }

  element = el + strlen (SLAD_NS);

  for (i=0; elements[i]; i++)
    if (!strcmp (element, elements[i]))
      break;

  if (elements[i])
    switch (i) {
    case ST_PLUGIN:
      XML_SetCharacterDataHandler (parser, nullhandler);
      XML_SetUserData (parser, NULL);
      plugin = calloc (sizeof (slad_plugin_t), 1);
      set = NULL;
      entry = NULL;
      if (!plugin)
	{
	  PRINTF("xml start(): could not calloc plugin\n");
	  return;
	}
      plugin->path = strdup (basename (dirname (config->pluginpath)));
      if (enqueue ((void ***) &(config->plugins), plugin)) {
	free (plugin);
	plugin = NULL;
	return;
      }
      plugin->name = strdup (getattr ("id", attr));
      break;
    case ST_PLUGINDESC:
      if (plugin) {
	XML_SetCharacterDataHandler (parser, stringhandler);
	XML_SetUserData (parser, &(plugin->desc));
      }
      break;
    case ST_PLUGINPATH:
      if (plugin) {
	XML_SetCharacterDataHandler (parser, stringhandler);
	XML_SetUserData (parser, &(plugin->path));
      }
      break;
    case ST_PLUGINSET:
      XML_SetCharacterDataHandler (parser, nullhandler);
      XML_SetUserData (parser, NULL);
      if (!plugin) {
	fprintf (stderr, "Plugin set encountered, but no plugin\n");
	return;
      }
      set = calloc (sizeof (slad_plugin_set_t), 1);
      entry = NULL;
      if (!set)
	{
	  PRINTF("xml start(): could not calloc set\n");
	  return;
	}
      if (enqueue ((void ***) &(plugin->sets), set)) {
	free (set);
	set = NULL;
	return;
      }
      set->name = strdup (getattr ("id", attr));
      break;
    case ST_PLUGINSETDESC:
      if (set) {
	XML_SetCharacterDataHandler (parser, stringhandler);
	XML_SetUserData (parser, &(set->desc));
      }
      break;
    case ST_PLUGINSETDEFAULT:
      if (set) {
	XML_SetCharacterDataHandler (parser, inthandler);
	XML_SetUserData (parser, &(set->is_default));
      }
      break;
    case ST_PLUGINSETENTRY:
      XML_SetCharacterDataHandler (parser, nullhandler);
      XML_SetUserData (parser, NULL);
      if (!set) {
	fprintf (stderr, "Plugin set entry encountered, but no plugin set\n");
	return;
      }
      entry = calloc (sizeof (slad_plugin_entry_t), 1);
      if (!entry)
	{
	  PRINTF("xml start(): could not calloc entry\n");
	  return;
	}
      if (enqueue ((void ***) &(set->entries), entry)) {
	free (entry);
	entry = NULL;
	return;
      }
      entry->name = strdup (getattr ("id", attr));
      break;
    case ST_PLUGINSETENTRYDESC:
      if (entry) {
	XML_SetCharacterDataHandler (parser, stringhandler);
	XML_SetUserData (parser, &(entry->desc));
      }
      break;
    case ST_PLUGINSETENTRYDEFAULT:
      if (entry) {
	XML_SetCharacterDataHandler (parser, inthandler);
	XML_SetUserData (parser, &(entry->is_default));
      }
      break;
    case ST_PLUGINSETENTRYBINARY:
      if (entry) {
	XML_SetCharacterDataHandler (parser, stringhandler);
	XML_SetUserData (parser, &(entry->binary));
      }
      break;
    case ST_PLUGINSETENTRYCMDLINE:
      if (entry) {
	XML_SetCharacterDataHandler (parser, stringhandler);
	XML_SetUserData (parser, &(entry->cmdline));
      }
      break;
    case ST_PLUGINSETENTRYMAXRUNTIME:
      if (entry) {
	XML_SetCharacterDataHandler (parser, inthandler);
	XML_SetUserData (parser, &(entry->max_runtime));
      }
      break;
    case ST_PLUGINSETENTRYMAXLOGLINES:
      if (entry) {
	XML_SetCharacterDataHandler (parser, inthandler);
	XML_SetUserData (parser, &(entry->max_loglines));
      }
      break;
    case ST_PLUGINSETENTRYRUNASUSER:
      if (entry) {
	XML_SetCharacterDataHandler (parser, stringhandler);
	XML_SetUserData (parser, &(entry->runasuser));
      }
      break;
    case ST_PLUGINSETENTRYRUNASGROUP:
      if (entry) {
	XML_SetCharacterDataHandler (parser, stringhandler);
	XML_SetUserData (parser, &(entry->runasgroup));
      }
      break;
    default:
      fprintf (stderr, "Unhandled element %s during XML parsing\n",
	       element);
    }
  else
    fprintf (stderr, "Unknown element %s during XML parsing\n", element);

  depth++;
}

static void XMLCALL end(void *data, const char *el) {
  depth--;
}

#define BUFSIZE 256

void load_plugin (const char * cfgname, sladd_config_t * config) {
  FILE * file;
  int len;
  int done = 0;
  XML_Parser parser;
  void * buffer;
  int result;
  char * tmp;

  parser = XML_ParserCreate (NULL);
  if (!parser)
    {
      PRINTF("could not create XML parser\n");
      return;
    }

  XML_SetElementHandler (parser, start, end);
  XML_SetCharacterDataHandler (parser, nullhandler);
  XML_UseParserAsHandlerArg (parser);
  XML_SetUserData (parser, config);

  if (!(file = fopen (cfgname, "r")))
    {
      PRINTF("could not open %s\n", cfgname);
      return;
    }

  tmp = config->pluginpath;
  config->pluginpath = (char *) cfgname;

  do {
    buffer = XML_GetBuffer (parser, BUFSIZE);
    if (!buffer) {
      fprintf (stderr, "Got no buffer for XML parsing: %s\n",
	       XML_ErrorString(XML_GetErrorCode(parser)));
      config->pluginpath = tmp;
      return;
    }

    if (fgets (buffer, BUFSIZE-1, file))
      len = strlen (buffer);
    else
      len = 0;
    done = feof (file);
    result = XML_Parse(parser, buffer, len, done);
    if (result == XML_STATUS_ERROR) {
      fprintf (stderr, "(%d) |%s|\n", len, (char *) buffer);
      fprintf (stderr, "Parse error at line %d: %s\n",
	       XML_GetCurrentLineNumber(parser),
	       XML_ErrorString(XML_GetErrorCode(parser)));
      config->pluginpath = tmp;
      return;
    }
  } while (!done);

  fclose (file);
  XML_ParserFree (parser);
  config->pluginpath = tmp;
}
